from jax import *
from jax.numpy import *


### FIXME import cycles would need to be resolved
# These submodules are separate because they are in an import cycle with
# jax and rely on the names imported above.
from jax import custom_derivatives as custom_derivatives
from jax import custom_batching as custom_batching
from jax import custom_transpose as custom_transpose
from jax import api_util as api_util
from jax import distributed as distributed
from jax import debug as debug
from jax import dlpack as dlpack
from jax import dtypes as dtypes
from jax import errors as errors
from jax import ffi as ffi
from jax import image as image
from jax import lax as lax
from jax import monitoring as monitoring
from jax import nn as nn
# from jax import numpy as numpy
from jax import ops as ops
from jax import profiler as profiler
# from jax import random as random # FIXME
from jax import scipy as scipy
from jax import sharding as sharding
from jax import stages as stages
from jax import tree_util as tree_util
from jax import util as util


def enable_performance_warnings():
    raise NotImplementedError


def print_jit_source(func, *args, **kwargs):
    raise NotImplementedError


# use global_rng from _random_jax module, so only 1 place
from jaxpy import random
from jaxpy._random_jax import _global_rng


# useful ops isto faria parte de outro ficheiro

# # not from import?
# from collections.abc import Callable, Sequence
# from jax._src.typing import (
#   Array, ArrayLike,
#   DType, DTypeLike, DeprecatedArg, DimSize, DuckTypedArray, Shape, StaticScalar,
# )
# import numpy as np
# def cat(xs: np.ndarray | Array | Sequence[ArrayLike],
#                 axis: int | None = 0, dtype: DTypeLike | None = None) -> Array:
#     return concatenate(xs, axis=axis, dtype=dtype)



import functools
import builtins


# not sure if should also allow function names, but want to make it such that it works
# even if jax deprecates suddently some function
def make_alias(func_name: str, new_name: str, namespace=None):
    if namespace is None:
        namespace = globals()
    
    # Look up the function by name in the namespace
    original_func = namespace.get(func_name) or getattr(builtins, func_name, None)
    
    #FIXME what todo if alias not available do nothing?
    if not callable(original_func):
        raise ValueError(f"{func_name!r} is not a callable in the given namespace.")
    
    @functools.wraps(original_func)
    def wrapper(*args, **kwargs):
        return original_func(*args, **kwargs)
    
    wrapper.__name__ = new_name
    wrapper.__qualname__ = new_name
    namespace[new_name] = wrapper
    return wrapper

make_alias("concatenate","cat")



### seed dev

# class _SingletonMeta(type):
#     _instances = {}
    
#     def __call__(cls, *args, **kwargs):
#         if cls not in cls._instances:
#             # This creates the actual instance
#             cls._instances[cls] = super().__call__(*args, **kwargs)
#         return cls._instances[cls]

# class _Singleton(metaclass=_SingletonMeta):
#     pass

# class _Global_RNG(_Singleton):
#     def __init__(self):
#         self.key = random.key(0)

#     def get_next(self):
#         self.key, new_key = random.split(self.key, 2)
        
#         return new_key
    
#     def set_seed(self, seed):
#         self.key = random.key(seed)

# _global_rng = _Global_RNG()



# def get_global_rng():
#     return _global_rng.key

# def set_seed(seed):
#     _global_rng.set_seed(seed)

# FIXME key...
def randn(shape,dtype=float,out_sharding=None, key=None):
    return random.normal(shape,dtype,out_sharding=out_sharding, key=key)

# # why like this?? make it more like numpy always... numpy is an api by now, will follow numpy by default...
def randint(minval,maxval,shape,dtype=int, out_sharding=None, key=None):
    return random.randint(shape, minval, maxval, dtype, out_sharding=out_sharding, key=key)


import jax


# vamos tentar dar replace do jit principalmente
# preciso tb perceber se grad precisa
from flax.nnx import grad, pmap, vmap
from flax.nnx import jit as _nnx_jit
from flax.nnx import Rngs

from jax import jit as _jax_jit
# vjg, jvp, hessian, not defined. usam jax por default. nao vou mudar

import typing as tp

# FIXME this makes no sense for now
def jit(
  fun: tp.Callable, /, *,
  in_shardings: tp.Any = None,
  out_shardings: tp.Any = None,
  static_argnums: int | tp.Sequence[int] | None = None,
  static_argnames: str | tp.Iterable[str] | None = None,
  donate_argnums: int | tp.Sequence[int] | None = None,
  donate_argnames: str | tp.Iterable[str] | None = None,
  keep_unused: bool = False,
  device: tp.Optional[jax.Device] | None = None,
  backend: str | None = None,
  inline: bool = False,
  abstracted_axes: tp.Any | None = None,
): #-> pjit.JitWrapped:
    '''
    jaxpy version, it preprocesses the function before sending it to jax
    '''

    # Check if the function is already a JIT function
    if hasattr(fun, 'jaxpy_jit'):
        # If it is, we can just return it
        return fun
    
    if not callable(fun):
        raise TypeError(f"Expected a callable, got {type(fun).__name__}")
    
    


    return _nnx_jit(fun, in_shardings=in_shardings, out_shardings=out_shardings,
                    static_argnums=static_argnums, static_argnames=static_argnames,
                    donate_argnums=donate_argnums, donate_argnames=donate_argnames,
                    keep_unused=keep_unused, device=device, backend=backend,
                    inline=inline, abstracted_axes=abstracted_axes)


######3 todo
'''

we need a lexer to process all globals e remover los dos returns puros.
lexer para todas as variaveis e funcoes usadas?

definir que novas variaveis dentro de bloco, ficaram locais excepto se for definido como return manualmente?
+ returnar tudo talvez seja uma boa opcao, visto que pode ser usado. Acho que e' mais simples. (menos eficiente again argumento, mas n muito em principio)
+ every assigment e' retornado

---------------------

Exemplo do que acharia um possivel erro de usar funcaoes
Non-jax function with dependencies? provokes a split on compiled graph in line:
previous dependence: b=jp.array([1,2,3])

previous line: b=jp.array([1,2,3])
-> line_nr: a = np.array([1,2,3])+b

further dependence: return a @ c


claro isto requer fazer track do uso das variaveis, e calls para funcoes not jittable
-----


a=2
d=None
e=None

@jp.jit
def impure():
    global e                                ## vars - e, global(e)     
    b=jp.array([1,2,3])                     ## FUNC - jp.array  ,assigment(b)                      
    temp=np.array([1,2,3])                  ## FUNC - np.array , assigment(temp) . suportamos todos os basic data types por isso ignoramos  
    temp2 = jax.array(np.array([1,2,3]))    ## FUNC - jax.array, np.array , assigment(temp2)
    b=b+temp                                ## vars - b, temp, assigment(b)
                                            ##
    c=a+b                                   ## vars - a, b, assigment(c)             
    print(c)                                ## vars - c, FUNC - print
                                            ##     
    d=a-b                                   ## vars - a, b, assigment(d)
    e=2*a-b                                 ## vars - a, b, assigment(e)
    return c                                ## return(c)     


# importante distinguir variavel na expressao a ser assign vs usada  (mas acho que e' simples, e' so todas as expressoes antes do =, confirmar com python syntax)

globals = [e]
func_args=[]
func_returns=[c]


variables=[a, np.array, jp.array, jax.array] # variáveis usadas antes de qualquer assignment 

in_variables = [a,] # variavel usada antes de qualquer assigment (mesmo que sejam incompatíveis, se não der para fazer fora Taíse error jax)
in_variables_static = [ np.array, jp.array, jax.array] #calables são automaticamente estaticos, ou mais que possamos definir manualmente

Proceed to jit callables first.





# transformed pure
def pure():
    global e

    temp=np.array([1,2,3])
    temp2 = jax.array(np.array([1,2,3])) # full expression out? so pk tem funcao nao jax, parece me bem

    @jp.jax.jit # algo assim
    def func(a,temp): #np array is supported as input, na verdade acho que devemos tentar sempre por arg, deixar jax dar erro se incorreto
        b=jp.array([1,2,3])
        b=b+temp
        c=a+b

        e=2*a-b
        return c, e

    # run jitted function
    c, e = func(a)
    
    # execute unjittable code
    print(c)

    return c


@jp.jit
def impure2():
    a=jp.array([1,2,3])
    b=jp.array([1,2,3])
    c=jp.array([1,2,3])
    temp=np.array(b)
    b=b+jp.array(temp)
    return b+a

# transformed pure
def pure():
    @jp.jax.jit 
    def func1(a,temp):
        a=jp.array([1,2,3])
        b=jp.array([1,2,3])
        c=jp.array([1,2,3]) # but dont return is not used
        return a, b

    a,b = func1()
    
    # execute unjittable code, raise WARNING
    temp=np.array(b)

    @jp.jax.jit 
    def func2(a, b, temp):
        b=b+jp.array(temp)
        return b+a
    
    outs = func2(a,b,temp)

    return outs



quero ter uma especie de funcao se jax jitable function. ou algo assim. um check inicial rapido, que por exemplo diz que jax functional calls sao jitable, 
funcoes python simples tambem mas que outras que chamem C code nao.

. get all variables used in function. Separar objectos, de funcoes, de classes (por enquanto nao sei se devia pensar em mais objectos)

comecar com apenas funcoes e a variaveis. coisas que tem que fazer
+ globals define valores que a jitted function deve retornar para dar set global

'''


import jaxpy as jp
import jaxpy.random as random

from jaxpy._preprocess import preprocess_function

# def test_simple():
#     seed = jp.Rngs(10)
#     seed_ = jp.Rngs(10)

#     def simple():
#         return random.normal((2, 3))
    
#     target = random.normal((2, 3), key=seed_.get_next())
    
#     f, info = preprocess_function(simple)
#     key_arg_name = info['key_arg_name']

#     test = f(**{key_arg_name: seed})

#     assert jp.all(target==test)

# def test_simple2():
#     seed = jp.Rngs(10)
#     seed_ = jp.Rngs(10)

#     def simple():
#         return random.normal((2, 3))
    
#     def other():
#         return simple()
    
#     target = random.normal((2, 3), key=seed_.get_next())

#     print(locals())
#     print(globals())

#     f, info = preprocess_function(other)
#     key_arg_name = info['key_arg_name']


#     test = f(**{key_arg_name: seed})

#     assert jp.all(target==test)

# def test_simple3():
#     '''
#     Ok so basically the problem is the following,
#     if we send other to preprocess, simple is not in scope,
#     so we can obtain the locals of the caller by inspect (obtain simple)

#     so when processing other, inner will not be processed, simple will,
#     and simple will search on globals+locals (should we join? yes for compilation)

#     '''
#     seed = jp.Rngs(10)
#     seed_ = jp.Rngs(10)

#     def simple():
#         return random.normal((2, 3))
    
#     def other():
#         def inner():
#             return simple()
        
#         return inner()
    
#     target = random.normal((2, 3), key=seed_.get_next())

#     f, info = preprocess_function(other)
#     key_arg_name = info['key_arg_name']

#     test = f(**{key_arg_name: seed})

#     assert jp.all(target==test)

def test_dynamic_defs():
    seed = jp.Rngs(10)
    seed_ = jp.Rngs(10)

    f_random=random.normal

    def other():
        def inner():
            b = (lambda: 1) if False else a
            return b()
        
        return inner()
    
    # in this case simple also needs the locals
    def simple():
        return f_random((2, 3))
    
    a = simple
    

    target = random.normal((2, 3), key=seed_.get_next())

    f, info = preprocess_function(other)
    key_arg_name = info['key_arg_name']

    test = f(**{key_arg_name: seed})

    assert jp.all(target==test)


# def test_simple_attr_error():
#     try:

#         def simple():
#             return random.normal_unexistant((2, 3))
        
#         f, info = preprocess_function(simple)
#         assert "key_arg_name" not in info

#         test = f()
#     except AttributeError as e:
#         assert "normal_unexistant" in str(e)

# def test_simple_attr_error_2():
#     try:

#         def simple():
#             return random.normal_unexistant((2, 3))
        
#         def other():
#             return simple()
        
#         f, info = preprocess_function(other)
#         assert "key_arg_name" not in info

#         test = f()
#     except AttributeError as e:
#         assert "normal_unexistant" in str(e)


# def test_simple_attr_error_3():
#     try:
#         def other():
#             def simple():
#                 return random.normal_unexistant((2, 3))
        

#             return simple()
        
#         f, info = preprocess_function(other)
#         assert "key_arg_name" not in info

#         test = f()
#     except AttributeError as e:
#         assert "normal_unexistant" in str(e)



# ## Class inside fixing..
# # this works
# def test_simple2_jit():
#     seed = jp.Rngs(10)
#     seed_ = jp.Rngs(20)

#     def simple():
#         return random.normal((2, 3), key=jp.Rngs(20).get_next())
    
#     target = random.normal((2, 3), key=seed_.get_next())
    
#     f = jp.jit(simple)

#     test = f()

#     assert jp.all(target==test)

def test_simple2():
    seed = jp.Rngs(10)
    seed_ = jp.Rngs(20)

    """
    This one is very intersting, because we do a call, and then
    atrr and call. We cannot preprocess the latter call, as we do not know the return 
    of previous call.

    The class thing is also interesting, because it works differently than function (no global scope)
    # it could help us know the attributes, but I need to think about it better.
    """

    def simple():
        # return random.normal((2, 3), key=jp.Rngs(20).get_next()) #
        k=jp.Rngs(20)
        return random.normal((2, 3), key=k()) # error
    
    target = random.normal((2, 3), key=seed_.get_next())
    
    f, info = preprocess_function(simple)
    key_arg_name = info['key_arg_name']

    test = f(**{key_arg_name: seed})

    assert jp.all(target==test)




import jax.random as _random_jax
import importlib
import inspect

overloads_str = ""
alls_list = []

def get_annotation(annotation):
    if annotation is inspect._empty:
        return 'Any'
    elif hasattr(annotation, '__name__'): # If it's a type, get its name
        return annotation.__name__
    else:
        return str(annotation)
    

def get_default(default):
    if default is inspect._empty:
        return inspect._empty
    elif hasattr(default, '__name__'): # If it's a type, get its name
        return default.__name__
    elif isinstance(default, str): # If it's a string, return it as is
        return f"'{default}'"
    else:
        return str(default)

# Iterate over all attributes in the module
for name in dir(_random_jax):
    attr = getattr(_random_jax, name)
    if callable(attr):  # Import only functions and callable object
        # if first argument is key, then we generate an overload
        sig = inspect.signature(attr)
        doc = attr.__doc__ or ""
        if len(sig.parameters) > 0 and list(sig.parameters.keys())[0] == 'key':
            # obtain parameters names, annotations and defaults
            params = []
            for k, a in list(sig.parameters.items())[1:]:  # Skip the first parameter (key)
                annotation = get_annotation(a.annotation)
                default = get_default(a.default)
                if default is inspect._empty:
                    default = ''
                else:
                    default = f"={default}"
                params.append(f"{k}: {annotation}{default}")

            # add key parameter at end as optional
            params.append("key: ArrayLike = None")
            
            # obtain return annotation
            return_annotation = get_annotation(sig.return_annotation)

            # create overload string
            overloads_str += "@overload\n"
            overloads_str += f"def {name}({', '.join(params)}) -> {return_annotation}: \n"
            overloads_str += f'  """{doc.strip()}"""\n'
            overloads_str += "  ...\n\n"
            alls_list.append(name)

print(overloads_str)

print(f"__all__ = {alls_list}")
            



        
        
        # globals()[name] = attr  # Inject into global scope


### import overloads, followed with dynamic implementation of functions
### we use separate file to avoid importing types
### overloads are generated by tools/generate_jax_random_overloads.py
### and are essential for static type analysis

from jax.random import * # first define all functions as in random
from jaxpy._random_jax_overloads import * # then overload with our definitions

import jax.random as _jax_random

import inspect as _inspect

# FIXME this needs to be acessible here, and singleton...
# FIXME we should be able to set the seed, and get the next key from jaxpy main
# jp._global_rng.__init__(0) would work so just make functions
from flax.nnx import Rngs as _Rngs
_global_rng = _Rngs(0)

# FIXME this could be usefull for front end user
# although for frontend we would just like to add __jaxpy_random_function__ to the function
# and maybe check if a key=None exists? I'm actually not sure
# another approach is just to do this, expect first argument to be key
def _jaxpy_random_wrapper_function(f, __module__ = "jaxpy.random_jax"):
    import functools
    # we need to remove the first parameter (key) from the signature
    sig = _inspect.signature(f)
    params = list(sig.parameters.values())

    if not params or params[0].name != 'key':
        raise ValueError(f"Function {f.__name__} does not have 'key' as first parameter.")

    # Remove 'key' from first position
    key_param = params.pop(0)
    # Re-add it as a keyword-only parameter with default=None
    key_param = key_param.replace(default=None, kind=_inspect.Parameter.KEYWORD_ONLY)
    params.append(key_param)

    new_sig = sig.replace(parameters=params)

    @functools.wraps(f)
    def wrapped(*args, key=None, **kwargs):
        if key is None:
            key = _global_rng.get_next()

        return f(key, *args, **kwargs)

    wrapped.__jaxpy_random_function__ = True
    wrapped.__signature__ = new_sig
    wrapped.__module__ = __module__ # set function module, could not make it work automatically, thus manual

    return wrapped






# all needs to be static for pylance to work properly
__all__ = []

# Iterate over all attributes in the module
for _name in dir(_jax_random):
    _attr = getattr(_jax_random, _name)
    if callable(_attr):
      # if first argument is key, we wrap the function with automatic key handling
      _sig = _inspect.signature(_attr)
      if len(_sig.parameters) > 0 and list(_sig.parameters.keys())[0] == 'key':
          _attr = _jaxpy_random_wrapper_function(_attr)
    # inject into global scope
    globals()[_name] = _attr
    __all__.append(_name)




import jax
import jax.numpy as jnp
from flax import nnx

import os

e=0
def non_pure_func(x):
    global e

    e+=1
    return (x*2).sum()


def test1():
    global e
    x = jnp.ones(2)

    e=0

    non_pure_func(x)
    assert e==1

    jax.vmap(non_pure_func)(x)
    assert e==2

    f2=jax.vmap(non_pure_func)
    f2(x)
    f2(x)

    assert e==4
    out = f2(jnp.ones(100))
    assert e==5 # chama apenas uma vez, nice
    assert out.shape == (100,)

    grad_f = jax.grad(non_pure_func)
    grad = grad_f(x)
    assert grad[0]==2
    assert e==6
    
    e=10
    grad_f(jnp.zeros(10))
    grad_f(x)
    assert e==12

    f_jit = jax.jit(non_pure_func)
    e=0
    f_jit(x)
    assert e==1
    f_jit(x)
    f_jit(x)
    assert e==1


class Jax_test_class:
    def __init__(self):
        self.x = jnp.ones(10)
        self.y = jnp.ones(10)
        self.z = 0
    
    def non_pure_func(self, x):
        self.z+=1
        self.x +=self.y

        return self.y.sum() + x

def f(test_class: Jax_test_class, x: jnp.array):
    x = test_class.non_pure_func(x)
    return x.sum()



def test_class_args():
    t = Jax_test_class()
    x=jnp.ones(10)
    f(t,x)
    assert t.z==1
    assert jnp.all(t.x == t.y*2)

    g=jax.grad(f,argnums=1)(t,x)
    assert t.z==2
    assert jnp.all(t.x == t.y*3)

    try:
        f_jit = jax.jit(f)
        f_jit(t,x)
        assert t.z==3
        assert jnp.all(t.x == t.y*4)
    except TypeError as e:
        # print(e)
        # problem on test class, only array arguments,
        # with non array marked as static args on jax.jit
        pass

    f_jit = jax.jit(f, static_argnums=0)
    f_jit(t,x)
    assert t.z==3
    # this gives error as it detects side effects
    # assert jnp.all(t.x == t.y*4)

    

def test_static_args():
    jax.config.update("jax_log_compiles", True)


    @jax.jit
    def f(x, use_relu: bool):
        return jax.nn.relu(x) if use_relu else x



    def f2(x, use_relu: bool):
        return jax.nn.relu(x) if use_relu else x

    def f3(x, use_relu: bool):
        return jax.nn.relu(x).sum() if use_relu else x.sum()

    # error because bool passa para jax array, que nao podemos fazer if..
    # print("try f")
    # x=jnp.array((10,-10,-1,1))
    # print(f(x,True))
    # print(f(x,False))

    print("try f2")
    f2=nnx.jit(f2,static_argnames=['use_relu'])

    x=jnp.array((10,-10,-1,1)).astype(float)
    print(f2(x,True))
    print(f2(x,False))
    print("try grad f2")
    grad_f2 =nnx.grad(nnx.jit( (f3),static_argnames=['use_relu']))
    print(grad_f2(x,True))
    print(grad_f2(x,False))


# class Dataloader(nnx.Module):
#     def __init__(self,data):
#         self.data=data
#         self.lenght = len(data)
#         self.count
#  I think we should not jit dataloaders, at least not in this way....
#  at most a pattern for scan?
#  but I mean, what is a dataloader, more than a list of a list with permutation indexes right?
#  this is jitable 

def test_classes():
    # this works great! it means dataloaders or other modules would work fine

    class Test(nnx.Module):
        def __init__(self):
            self.a = 1
            self.b = jnp.array(1)

        def __call__(self):
            self.a+=1
            self.b +=1

            return self.a, self.b

    t=Test()
    print(t())
    print(t())
    print(t())

    def jit_test(c):
        return c(),c(),c()

    print(jit_test(t))

    f = nnx.jit(jit_test)
    print(jit_test(t))
    print(jit_test(t))
    print(jit_test(t))
    print(jit_test(t))
    print(jit_test(t))

    print(t)



# FIXME error since RngStream comes from another context
# jp.grad does not like this, works well with jax key
# def test_grad_with_random_eager():
#     import jaxpy as jp

#     print(jp.randn((10,1)))
#     print("now build test")
#     def test():
#         return jp.sum(jp.randn((10,10)) @ jp.randn((10,1)))

#     print(test())
#     print(test())

#     print("now jit")

#     #fails
#     # jtest = jp.jit(test)

#     # print(jtest())
#     # print(jtest())

#     def test2(x):
#         return jp.sum(x @ jp.randn((10,1)))

#     print("other version")

#     print(test2(jp.ones((10,10))))
#     print(test2(jp.ones((10,10))))

#     # works fine what is nice. so just jit!
#     grad_test = jp.grad(test2)
#     print(grad_test(jp.ones((10,10))))
#     print(grad_test(jp.ones((10,10))))


# nnx.Module

# nnx.jit


# from flax.nnx import Rngs

# # _global_rng = Rngs(0)


# class _SingletonMeta(type):
#     _instances = {}
    
#     def __call__(cls, *args, **kwargs):
#         if cls not in cls._instances:
#             # This creates the actual instance
#             cls._instances[cls] = super().__call__(*args, **kwargs)
#         return cls._instances[cls]

# class _Singleton(metaclass=_SingletonMeta):
#     pass

# class _Global_RNG(_Singleton):
#     def __init__(self):
#         self.key = Rngs(0)

#     def get_next(self):
#         return self.key.get_next()
    
#     def set_seed(self, seed):
#         self.key = Rngs(seed)

# _global_rng = _Global_RNG()



from typing import overload

from typing import Callable, Any
from collections.abc import Sequence
from jax._src.typing import ArrayLike, DTypeLike, Array
RealArray = ArrayLike
IntegerArray = ArrayLike
# TODO: Import or define these to match
# https://github.com/numpy/numpy/blob/main/numpy/typing/_dtype_like.py.
DTypeLikeInt = DTypeLike
DTypeLikeUInt = DTypeLike
DTypeLikeFloat = DTypeLike
Shape = Sequence[int]

### overloads from tools/generate_jax_random_overloads.py
### Manually edited the docstrings to put key optional
@overload
def ball(d: int, p: float=2, shape: Shape=(), dtype: DTypeLikeFloat=float, key: ArrayLike = None) -> Any: 
  """Sample uniformly from the unit Lp ball.

  Reference: https://arxiv.org/abs/math/0503650.

  Args:
    d: a nonnegative int representing the dimensionality of the ball.
    p: a float representing the p parameter of the Lp norm.
    shape: optional, the batch dimensions of the result. Default ().
    dtype: optional, a float dtype for the returned values (default float64 if
      jax_enable_x64 is true, otherwise float32).
    key: optional, a PRNG key used as the random key.

  Returns:
    A random array of shape `(*shape, d)` and specified dtype."""
  ...

@overload
def bernoulli(p: RealArray=0.5, shape: Shape | None=None, mode: str='low', key: ArrayLike = None) -> Array: 
  """Sample Bernoulli random values with given shape and mean.

  The values are distributed according to the probability mass function:

  .. math::
     f(k; p) = p^k(1 - p)^{1 - k}

  where :math:`k \in \{0, 1\}` and :math:`0 \le p \le 1`.

  Args:
    p: optional, a float or array of floats for the mean of the random
      variables. Must be broadcast-compatible with ``shape``. Default 0.5.
    shape: optional, a tuple of nonnegative integers representing the result
      shape. Must be broadcast-compatible with ``p.shape``. The default (None)
      produces a result shape equal to ``p.shape``.
    mode: optional, "high" or "low" for how many bits to use when sampling.
      default='low'. Set to "high" for correct sampling at small values of
      `p`. When sampling in float32, bernoulli samples with mode='low' produce
      incorrect results for p < ~1E-7. mode="high" approximately doubles the
      cost of sampling.
    key: optional, a PRNG key used as the random key.

  Returns:
    A random array with boolean dtype and shape given by ``shape`` if ``shape``
    is not None, or else ``p.shape``."""
  ...

@overload
def beta(a: RealArray, b: RealArray, shape: Shape | None=None, dtype: DTypeLikeFloat=float, key: ArrayLike = None) -> Array: 
  """Sample Beta random values with given shape and float dtype.

  The values are distributed according to the probability density function:

  .. math::
     f(x;a,b) \propto x^{a - 1}(1 - x)^{b - 1}

  on the domain :math:`0 \le x \le 1`.

  Args:
    a: a float or array of floats broadcast-compatible with ``shape``
      representing the first parameter "alpha".
    b: a float or array of floats broadcast-compatible with ``shape``
      representing the second parameter "beta".
    shape: optional, a tuple of nonnegative integers specifying the result
      shape. Must be broadcast-compatible with ``a`` and ``b``. The default
      (None) produces a result shape by broadcasting ``a`` and ``b``.
    dtype: optional, a float dtype for the returned values (default float64 if
      jax_enable_x64 is true, otherwise float32).
    key: optional, a PRNG key used as the random key.

  Returns:
    A random array with the specified dtype and shape given by ``shape`` if
    ``shape`` is not None, or else by broadcasting ``a`` and ``b``."""
  ...

@overload
def binomial(n: RealArray, p: RealArray, shape: Shape | None=None, dtype: DTypeLikeFloat=float, key: ArrayLike = None) -> Array: 
  """Sample Binomial random values with given shape and float dtype.

  The values are returned according to the probability mass function:

  .. math::
      f(k;n,p) = \binom{n}{k}p^k(1-p)^{n-k}

  on the domain :math:`0 < p < 1`, and where :math:`n` is a nonnegative integer
  representing the number of trials and :math:`p` is a float representing the
  probability of success of an individual trial.

  Args:
    n: a float or array of floats broadcast-compatible with ``shape``
      representing the number of trials.
    p: a float or array of floats broadcast-compatible with ``shape``
      representing the probability of success of an individual trial.
    shape: optional, a tuple of nonnegative integers specifying the result
      shape. Must be broadcast-compatible with ``n`` and ``p``.
      The default (None) produces a result shape equal to ``np.broadcast(n, p).shape``.
    dtype: optional, a float dtype for the returned values (default float64 if
      jax_enable_x64 is true, otherwise float32).
    key: optional, a PRNG key used as the random key.

  Returns:
    A random array with the specified dtype and with shape given by
    ``np.broadcast(n, p).shape``."""
  ...

@overload
def bits(shape: Shape=(), dtype: DTypeLikeUInt | None=None, out_sharding: Any=None, key: ArrayLike = None) -> Array: 
  """Sample uniform bits in the form of unsigned integers.

  Args:
    shape: optional, a tuple of nonnegative integers representing the result
      shape. Default ``()``.
    dtype: optional, an unsigned integer dtype for the returned values (default
      ``uint64`` if ``jax_enable_x64`` is true, otherwise ``uint32``).
    key: optional, a PRNG key used as the random key.

  Returns:
    A random array with the specified shape and dtype."""
  ...

@overload
def categorical(logits: RealArray, axis: int=-1, shape: Shape | None=None, replace: bool=True, mode: str | None=None, key: ArrayLike = None) -> Array: 
  """Sample random values from categorical distributions.

  Sampling with replacement uses the Gumbel max trick. Sampling without replacement uses
  the Gumbel top-k trick. See [1] for reference.

  Args:
    logits: Unnormalized log probabilities of the categorical distribution(s) to sample from,
      so that `softmax(logits, axis)` gives the corresponding probabilities.
    axis: Axis along which logits belong to the same categorical distribution.
    shape: Optional, a tuple of nonnegative integers representing the result shape.
      Must be broadcast-compatible with ``np.delete(logits.shape, axis)``.
      The default (None) produces a result shape equal to ``np.delete(logits.shape, axis)``.
    replace: If True (default), perform sampling with replacement. If False, perform
      sampling without replacement.
    mode: optional, "high" or "low" for how many bits to use in the gumbel sampler.
      The default is determined by the ``use_high_dynamic_range_gumbel`` config,
      which defaults to "low". With mode="low", in float32 sampling will be biased
      for events with probability less than about 1E-7; with mode="high" this limit
      is pushed down to about 1E-14. mode="high" approximately doubles the cost of
      sampling.
    key: optional, a PRNG key used as the random key.

  Returns:
    A random array with int dtype and shape given by ``shape`` if ``shape``
    is not None, or else ``np.delete(logits.shape, axis)``.

  References:
    .. [1] Wouter Kool, Herke van Hoof, Max Welling. "Stochastic Beams and Where to Find
      Them: The Gumbel-Top-k Trick for Sampling Sequences Without Replacement".
      Proceedings of the 36th International Conference on Machine Learning, PMLR
      97:3499-3508, 2019. https://proceedings.mlr.press/v97/kool19a.html."""
  ...

@overload
def cauchy(shape: Shape=(), dtype: DTypeLikeFloat=float, key: ArrayLike = None) -> Array: 
  """Sample Cauchy random values with given shape and float dtype.

  The values are distributed according to the probability density function:

  .. math::
     f(x) \propto \frac{1}{x^2 + 1}

  on the domain :math:`-\infty < x < \infty`

  Args:
    shape: optional, a tuple of nonnegative integers representing the result
      shape. Default ().
    dtype: optional, a float dtype for the returned values (default float64 if
      jax_enable_x64 is true, otherwise float32).
    key: optional, a PRNG key used as the random key.

  Returns:
    A random array with the specified shape and dtype."""
  ...

@overload
def chisquare(df: RealArray, shape: Shape | None=None, dtype: DTypeLikeFloat=float, key: ArrayLike = None) -> Array: 
  """Sample Chisquare random values with given shape and float dtype.

  The values are distributed according to the probability density function:

  .. math::
     f(x; \nu) \propto x^{\nu/2 - 1}e^{-x/2}

  on the domain :math:`0 < x < \infty`, where :math:`\nu > 0` represents the
  degrees of freedom, given by the parameter ``df``.

  Args:
    df: a float or array of floats broadcast-compatible with ``shape``
      representing the parameter of the distribution.
    shape: optional, a tuple of nonnegative integers specifying the result
      shape. Must be broadcast-compatible with ``df``. The default (None)
      produces a result shape equal to ``df.shape``.
    dtype: optional, a float dtype for the returned values (default float64 if
      jax_enable_x64 is true, otherwise float32).
    key: optional, a PRNG key used as the random key.

  Returns:
    A random array with the specified dtype and with shape given by ``shape`` if
    ``shape`` is not None, or else by ``df.shape``."""
  ...

@overload
def choice(a: int | ArrayLike, shape: Shape=(), replace: bool=True, p: RealArray | None=None, axis: int=0, mode: str | None=None, key: ArrayLike = None) -> Array: 
  """Generates a random sample from a given array.

  .. warning::
    If ``p`` has fewer non-zero elements than the requested number of samples,
    as specified in ``shape``, and ``replace=False``, the output of this
    function is ill-defined. Please make sure to use appropriate inputs.

  Args:
    a : array or int. If an ndarray, a random sample is generated from
      its elements. If an int, the random sample is generated as if a were
      arange(a).
    shape : tuple of ints, optional. Output shape.  If the given shape is,
      e.g., ``(m, n)``, then ``m * n`` samples are drawn.  Default is (),
      in which case a single value is returned.
    replace : boolean.  Whether the sample is with or without replacement.
      Default is True.
    p : 1-D array-like, The probabilities associated with each entry in a.
      If not given the sample assumes a uniform distribution over all
      entries in a.
    axis: int, optional. The axis along which the selection is performed.
      The default, 0, selects by row.
    mode: optional, "high" or "low" for how many bits to use in the gumbel sampler
      when `p is None` and `replace = False`. The default is determined by the
      ``use_high_dynamic_range_gumbel`` config, which defaults to "low". With mode="low",
      in float32 sampling will be biased for choices with probability less than about
      1E-7; with mode="high" this limit is pushed down to about 1E-14. mode="high"
      approximately doubles the cost of sampling.
    key: optional, a PRNG key used as the random key.

  Returns:
    An array of shape `shape` containing samples from `a`."""
  ...

@overload
def clone(key: ArrayLike = None) -> Any: 
  """Clone a key for reuse

  Outside the context of key reuse checking (see :mod:`jax.experimental.key_reuse`)
  this function operates as an identity.

  Examples:

    >>> import jax
    >>> key = jax.random.key(0)
    >>> data = jax.random.uniform(key)
    >>> cloned_key = jax.random.clone(key)
    >>> same_data = jax.random.uniform(cloned_key)
    >>> assert data == same_data"""
  ...

@overload
def dirichlet(alpha: RealArray, shape: Shape | None=None, dtype: DTypeLikeFloat=float, key: ArrayLike = None) -> Array: 
  """Sample Dirichlet random values with given shape and float dtype.

  The values are distributed according to the probability density function:

  .. math::
     f(\{x_i\}; \{\alpha_i\}) \propto \prod_{i=1}^k x_i^{\alpha_i - 1}

  Where :math:`k` is the dimension, and :math:`\{x_i\}` satisfies

  .. math::
     \sum_{i=1}^k x_i = 1

  and :math:`0 \le x_i \le 1` for all :math:`x_i`.

  Args:
    alpha: an array of shape ``(..., n)`` used as the concentration
      parameter of the random variables.
    shape: optional, a tuple of nonnegative integers specifying the result
      batch shape; that is, the prefix of the result shape excluding the last
      element of value ``n``. Must be broadcast-compatible with
      ``alpha.shape[:-1]``. The default (None) produces a result shape equal to
      ``alpha.shape``.
    dtype: optional, a float dtype for the returned values (default float64 if
      jax_enable_x64 is true, otherwise float32).
    key: optional, a PRNG key used as the random key.

  Returns:
    A random array with the specified dtype and shape given by
    ``shape + (alpha.shape[-1],)`` if ``shape`` is not None, or else
    ``alpha.shape``."""
  ...

@overload
def double_sided_maxwell(loc: RealArray, scale: RealArray, shape: Shape=(), dtype: DTypeLikeFloat=float, key: ArrayLike = None) -> Array: 
  """Sample from a double sided Maxwell distribution.

  The values are distributed according to the probability density function:

  .. math::
     f(x;\mu,\sigma) \propto z^2 e^{-z^2 / 2}

  where :math:`z = (x - \mu) / \sigma`, with the center :math:`\mu` specified by
  ``loc`` and the scale :math:`\sigma` specified by ``scale``.

  Args:
    loc: The location parameter of the distribution.
    scale: The scale parameter of the distribution.
    shape: The shape added to the parameters loc and scale broadcastable shape.
    dtype: The type used for samples.
    key: optional, a PRNG key used as the random key.

  Returns:
    A jnp.array of samples."""
  ...

@overload
def exponential(shape: Shape=(), dtype: DTypeLikeFloat=float, key: ArrayLike = None) -> Array: 
  """Sample Exponential random values with given shape and float dtype.

  The values are distributed according to the probability density function:

  .. math::
     f(x) = e^{-x}

  on the domain :math:`0 \le x < \infty`.

  Args:
    shape: optional, a tuple of nonnegative integers representing the result
      shape. Default ().
    dtype: optional, a float dtype for the returned values (default float64 if
      jax_enable_x64 is true, otherwise float32).
    key: optional, a PRNG key used as the random key.

  Returns:
    A random array with the specified shape and dtype."""
  ...

@overload
def f(dfnum: RealArray, dfden: RealArray, shape: Shape | None=None, dtype: DTypeLikeFloat=float, key: ArrayLike = None) -> Array: 
  """Sample F-distribution random values with given shape and float dtype.

  The values are distributed according to the probability density function:

  .. math::
     f(x; \nu_1, \nu_2) \propto x^{\nu_1/2 - 1}\left(1 + \frac{\nu_1}{\nu_2}x\right)^{
      -(\nu_1 + \nu_2) / 2}

  on the domain :math:`0 < x < \infty`. Here :math:`\nu_1` is the degrees of
  freedom of the numerator (``dfnum``), and :math:`\nu_2` is the degrees of
  freedom of the denominator (``dfden``).

  Args:
    dfnum: a float or array of floats broadcast-compatible with ``shape``
      representing the numerator's ``df`` of the distribution.
    dfden: a float or array of floats broadcast-compatible with ``shape``
      representing the denominator's ``df`` of the distribution.
    shape: optional, a tuple of nonnegative integers specifying the result
      shape. Must be broadcast-compatible with ``dfnum`` and ``dfden``.
      The default (None) produces a result shape equal to ``dfnum.shape``,
      and ``dfden.shape``.
    dtype: optional, a float dtype for the returned values (default float64 if
      jax_enable_x64 is true, otherwise float32).
    key: optional, a PRNG key used as the random key.

  Returns:
    A random array with the specified dtype and with shape given by ``shape`` if
    ``shape`` is not None, or else by ``df.shape``."""
  ...

@overload
def fold_in(data: IntegerArray, key: ArrayLike = None) -> Array: 
  """Folds in data to a PRNG key to form a new PRNG key.

  Args:
    data: a 32-bit integer representing data to be folded into the key.
    key: optional, a PRNG key (from ``key``, ``split``, ``fold_in``).

  Returns:
    A new PRNG key that is a deterministic function of the inputs and is
    statistically safe for producing a stream of new pseudo-random values."""
  ...

@overload
def gamma(a: RealArray, shape: Shape | None=None, dtype: DTypeLikeFloat=float, key: ArrayLike = None) -> Array: 
  """Sample Gamma random values with given shape and float dtype.

  The values are distributed according to the probability density function:

  .. math::
     f(x;a) \propto x^{a - 1} e^{-x}

  on the domain :math:`0 \le x < \infty`, with :math:`a > 0`.

  This is the standard gamma density, with a unit scale/rate parameter.
  Dividing the sample output by the rate is equivalent to sampling from
  *gamma(a, rate)*, and multiplying the sample output by the scale is equivalent
  to sampling from *gamma(a, scale)*.

  Args:
    a: a float or array of floats broadcast-compatible with ``shape``
      representing the parameter of the distribution.
    shape: optional, a tuple of nonnegative integers specifying the result
      shape. Must be broadcast-compatible with ``a``. The default (None)
      produces a result shape equal to ``a.shape``.
    dtype: optional, a float dtype for the returned values (default float64 if
      jax_enable_x64 is true, otherwise float32).
    key: optional, a PRNG key used as the random key.

  Returns:
    A random array with the specified dtype and with shape given by ``shape`` if
    ``shape`` is not None, or else by ``a.shape``.

  See Also:
    loggamma : sample gamma values in log-space, which can provide improved
      accuracy for small values of ``a``."""
  ...

@overload
def generalized_normal(p: float, shape: Shape=(), dtype: DTypeLikeFloat=float, key: ArrayLike = None) -> Array: 
  """Sample from the generalized normal distribution.

  The values are returned according to the probability density function:

  .. math::
     f(x;p) \propto e^{-|x|^p}

  on the domain :math:`-\infty < x < \infty`, where :math:`p > 0` is the
  shape parameter.

  Args:
    p: a float representing the shape parameter.
    shape: optional, the batch dimensions of the result. Default ().
    dtype: optional, a float dtype for the returned values (default float64 if
      jax_enable_x64 is true, otherwise float32).
    key: optional, a PRNG key used as the random key.

  Returns:
    A random array with the specified shape and dtype."""
  ...

@overload
def geometric(p: RealArray, shape: Shape | None=None, dtype: DTypeLikeInt=int, key: ArrayLike = None) -> Array: 
  """Sample Geometric random values with given shape and float dtype.

  The values are returned according to the probability mass function:

  .. math::
      f(k;p) = p(1-p)^{k-1}

  on the domain :math:`0 < p < 1`.

  Args:
    p: a float or array of floats broadcast-compatible with ``shape``
      representing the probability of success of an individual trial.
    shape: optional, a tuple of nonnegative integers specifying the result
      shape. Must be broadcast-compatible with ``p``. The default
      (None) produces a result shape equal to ``np.shape(p)``.
    dtype: optional, a int dtype for the returned values (default int64 if
      jax_enable_x64 is true, otherwise int32).
    key: optional, a PRNG key used as the random key.

  Returns:
    A random array with the specified dtype and with shape given by ``shape`` if
    ``shape`` is not None, or else by ``p.shape``."""
  ...

@overload
def gumbel(shape: Shape=(), dtype: DTypeLikeFloat=float, mode: str | None=None, key: ArrayLike = None) -> Array: 
  """Sample Gumbel random values with given shape and float dtype.

  The values are distributed according to the probability density function:

  .. math::
     f(x) = e^{-(x + e^{-x})}

  Args:
    shape: optional, a tuple of nonnegative integers representing the result
      shape. Default ().
    dtype: optional, a float dtype for the returned values (default float64 if
      jax_enable_x64 is true, otherwise float32).
    mode: optional, "high" or "low" for how many bits to use when sampling.
      The default is determined by the ``use_high_dynamic_range_gumbel`` config,
      which defaults to "low". When drawing float32 samples, with mode="low" the
      uniform resolution is such that the largest possible gumbel logit is ~16;
      with mode="high" this is increased to ~32, at approximately double the
      computational cost.
    key: optional, a PRNG key used as the random key.

  Returns:
    A random array with the specified shape and dtype."""
  ...

@overload
def laplace(shape: Shape=(), dtype: DTypeLikeFloat=float, key: ArrayLike = None) -> Array: 
  """Sample Laplace random values with given shape and float dtype.

  The values are distributed according to the probability density function:

  .. math::
    f(x) = \frac{1}{2}e^{-|x|}

  Args:
    shape: optional, a tuple of nonnegative integers representing the result
      shape. Default ().
    dtype: optional, a float dtype for the returned values (default float64 if
      jax_enable_x64 is true, otherwise float32).
    key: optional, a PRNG key used as the random key.

  Returns:
    A random array with the specified shape and dtype."""
  ...

@overload
def loggamma(a: RealArray, shape: Shape | None=None, dtype: DTypeLikeFloat=float, key: ArrayLike = None) -> Array: 
  """Sample log-gamma random values with given shape and float dtype.

  This function is implemented such that the following will hold for a
  dtype-appropriate tolerance::

    np.testing.assert_allclose(jnp.exp(loggamma(*args)), gamma(*args), rtol=rtol)

  The benefit of log-gamma is that for samples very close to zero (which occur frequently
  when `a << 1`) sampling in log space provides better precision.

  Args:
    a: a float or array of floats broadcast-compatible with ``shape``
      representing the parameter of the distribution.
    shape: optional, a tuple of nonnegative integers specifying the result
      shape. Must be broadcast-compatible with ``a``. The default (None)
      produces a result shape equal to ``a.shape``.
    dtype: optional, a float dtype for the returned values (default float64 if
      jax_enable_x64 is true, otherwise float32).
    key: optional, a PRNG key used as the random key.

  Returns:
    A random array with the specified dtype and with shape given by ``shape`` if
    ``shape`` is not None, or else by ``a.shape``.

  See Also:
    gamma : standard gamma sampler."""
  ...

@overload
def logistic(shape: Shape=(), dtype: DTypeLikeFloat=float, key: ArrayLike = None) -> Array: 
  """Sample logistic random values with given shape and float dtype.

  The values are distributed according to the probability density function:

  .. math::
     f(x) = \frac{e^{-x}}{(1 + e^{-x})^2}

  Args:
    shape: optional, a tuple of nonnegative integers representing the result
      shape. Default ().
    dtype: optional, a float dtype for the returned values (default float64 if
      jax_enable_x64 is true, otherwise float32).
    key: optional, a PRNG key used as the random key.

  Returns:
    A random array with the specified shape and dtype."""
  ...

@overload
def lognormal(sigma: RealArray=1.0, shape: Shape | None=None, dtype: DTypeLikeFloat=float, key: ArrayLike = None) -> Array: 
  """Sample lognormal random values with given shape and float dtype.

  The values are distributed according to the probability density function:

  .. math::
      f(x) = \frac{1}{x\sqrt{2\pi\sigma^2}}\exp\left(-\frac{(\log x)^2}{2\sigma^2}\right)

  on the domain :math:`x > 0`.

  Args:
    sigma: a float or array of floats broadcast-compatible with ``shape`` representing
      the standard deviation of the underlying normal distribution. Default 1.
    shape: optional, a tuple of nonnegative integers specifying the result
      shape. The default (None) produces a result shape equal to ``()``.
    dtype: optional, a float dtype for the returned values (default float64 if
      jax_enable_x64 is true, otherwise float32).
    key: optional, a PRNG key used as the random key.

  Returns:
    A random array with the specified dtype and with shape given by ``shape``."""
  ...

@overload
def maxwell(shape: Shape=(), dtype: DTypeLikeFloat=float, key: ArrayLike = None) -> Array: 
  """Sample from a one sided Maxwell distribution.

  The values are distributed according to the probability density function:

  .. math::
     f(x) \propto x^2 e^{-x^2 / 2}

  on the domain :math:`0 \le x < \infty`.

  Args:
    shape: The shape of the returned samples.
    dtype: The type used for samples.
    key: optional, a PRNG key.

  Returns:
    A jnp.array of samples, of shape `shape`."""
  ...

@overload
def multinomial(n: RealArray, p: RealArray, shape: Shape | None=None, dtype: DTypeLikeFloat=float, unroll: int | bool=1, key: ArrayLike = None) -> Any: 
  """Sample from a multinomial distribution.

  The probability mass function is

  .. math::
      f(x;n,p) = \frac{n!}{x_1! \ldots x_k!} p_1^{x_1} \ldots p_k^{x_k}

  Args:
    n: number of trials. Should have shape broadcastable to ``p.shape[:-1]``.
    p: probability of each outcome, with outcomes along the last axis.
    shape: optional, a tuple of nonnegative integers specifying the result batch
      shape, that is, the prefix of the result shape excluding the last axis.
      Must be broadcast-compatible with ``p.shape[:-1]``. The default (None)
      produces a result shape equal to ``p.shape``.
    dtype: optional, a float dtype for the returned values (default float64 if
      jax_enable_x64 is true, otherwise float32).
    unroll: optional, unroll parameter passed to :func:`jax.lax.scan` inside the
      implementation of this function.
    key: optional, a PRNG key used as the random key.

  Returns:
    An array of counts for each outcome with the specified dtype and with shape
      ``p.shape`` if ``shape`` is None, otherwise ``shape + (p.shape[-1],)``."""
  ...

@overload
def multivariate_normal(mean: RealArray, cov: RealArray, shape: Shape | None=None, dtype: DTypeLikeFloat | None=None, method: str='cholesky', key: ArrayLike = None) -> Array: 
  """Sample multivariate normal random values with given mean and covariance.

  The values are returned according to the probability density function:

  .. math::
     f(x;\mu, \Sigma) = (2\pi)^{-k/2} \det(\Sigma)^{-1}e^{-\frac{1}{2}(x - \mu)^T \Sigma^{-1} (x - \mu)}

  where :math:`k` is the dimension, :math:`\mu` is the mean (given by ``mean``) and
  :math:`\Sigma` is the covariance matrix (given by ``cov``).

  Args:
    mean: a mean vector of shape ``(..., n)``.
    cov: a positive definite covariance matrix of shape ``(..., n, n)``. The
      batch shape ``...`` must be broadcast-compatible with that of ``mean``.
    shape: optional, a tuple of nonnegative integers specifying the result
      batch shape; that is, the prefix of the result shape excluding the last
      axis. Must be broadcast-compatible with ``mean.shape[:-1]`` and
      ``cov.shape[:-2]``. The default (None) produces a result batch shape by
      broadcasting together the batch shapes of ``mean`` and ``cov``.
    dtype: optional, a float dtype for the returned values (default float64 if
      jax_enable_x64 is true, otherwise float32).
    method: optional, a method to compute the factor of ``cov``.
      Must be one of 'svd', 'eigh', and 'cholesky'. Default 'cholesky'. For
      singular covariance matrices, use 'svd' or 'eigh'.
    key: optional, a PRNG key used as the random key.
  Returns:
    A random array with the specified dtype and shape given by
    ``shape + mean.shape[-1:]`` if ``shape`` is not None, or else
    ``broadcast_shapes(mean.shape[:-1], cov.shape[:-2]) + mean.shape[-1:]``."""
  ...

@overload
def normal(shape: Shape=(), dtype: DTypeLikeFloat=float, out_sharding: Any=None, key: ArrayLike = None) -> Array: 
  """Sample standard normal random values with given shape and float dtype.

  The values are returned according to the probability density function:

  .. math::
     f(x) = \frac{1}{\sqrt{2\pi}}e^{-x^2/2}

  on the domain :math:`-\infty < x < \infty`

  Args:
    shape: optional, a tuple of nonnegative integers representing the result
      shape. Default ().
    dtype: optional, a float dtype for the returned values (default float64 if
      jax_enable_x64 is true, otherwise float32).
    key: optional, a PRNG key used as the random key.

  Returns:
    A random array with the specified shape and dtype."""
  ...

@overload
def orthogonal(n: int, shape: Shape=(), dtype: DTypeLikeFloat=float, m: int | None=None, key: ArrayLike = None) -> Array: 
  """Sample uniformly from the orthogonal group O(n).

  If the dtype is complex, sample uniformly from the unitary group U(n).

  For unequal rows and columns, this samples a semi-orthogonal matrix instead.
  That is, if :math:`A` is the resulting matrix and :math:`A^*` is its conjugate
  transpose, then:

  - If :math:`n \leq m`, the rows are mutually orthonormal: :math:`A A^* = I_n`.
  - If :math:`m \leq n`, the columns are mutually orthonormal: :math:`A^* A = I_m`.

  Args:
    n: an integer indicating the number of rows.
    shape: optional, the batch dimensions of the result. Default ().
    dtype: optional, a float dtype for the returned values (default float64 if
      jax_enable_x64 is true, otherwise float32).
    m: an integer indicating the number of columns. Defaults to `n`.
    key: optional, a PRNG key used as the random key.

  Returns:
    A random array of shape `(*shape, n, m)` and specified dtype.

  References:
    .. [1] Mezzadri, Francesco. (2007). "How to generate random matrices from
           the classical compact groups". Notices of the American Mathematical
           Society, 54(5), 592-604. https://arxiv.org/abs/math-ph/0609050."""
  ...

@overload
def pareto(b: RealArray, shape: Shape | None=None, dtype: DTypeLikeFloat=float, key: ArrayLike = None) -> Array: 
  """Sample Pareto random values with given shape and float dtype.

  The values are distributed according to the probability density function:

  .. math::
     f(x; b) = b / x^{b + 1}

  on the domain :math:`1 \le x < \infty` with :math:`b > 0`

  Args:
    b: a float or array of floats broadcast-compatible with ``shape``
      representing the parameter of the distribution.
    shape: optional, a tuple of nonnegative integers specifying the result
      shape. Must be broadcast-compatible with ``b``. The default (None)
      produces a result shape equal to ``b.shape``.
    dtype: optional, a float dtype for the returned values (default float64 if
      jax_enable_x64 is true, otherwise float32).
    key: optional, a PRNG key used as the random key.

  Returns:
    A random array with the specified dtype and with shape given by ``shape`` if
    ``shape`` is not None, or else by ``b.shape``."""
  ...

@overload
def permutation(x: int | ArrayLike, axis: int=0, independent: bool=False, out_sharding: Any=None, key: ArrayLike = None) -> Array: 
  """Returns a randomly permuted array or range.

  Args:
    x: int or array. If x is an integer, randomly shuffle np.arange(x).
      If x is an array, randomly shuffle its elements.
    axis: int, optional. The axis which x is shuffled along. Default is 0.
    independent: bool, optional. If set to True, each individual vector along
      the given axis is shuffled independently. Default is False.
    key: optional, a PRNG key used as the random key.

  Returns:
    A shuffled version of x or array range"""
  ...

@overload
def poisson(lam: RealArray, shape: Shape | None=None, dtype: DTypeLikeInt=int, key: ArrayLike = None) -> Array: 
  """Sample Poisson random values with given shape and integer dtype.

  The values are distributed according to the probability mass function:

  .. math::
     f(k; \lambda) = \frac{\lambda^k e^{-\lambda}}{k!}

  Where `k` is a non-negative integer and :math:`\lambda > 0`.

  Args:
    lam: rate parameter (mean of the distribution), must be >= 0. Must be broadcast-compatible with ``shape``
    shape: optional, a tuple of nonnegative integers representing the result
      shape. Default (None) produces a result shape equal to ``lam.shape``.
    dtype: optional, a integer dtype for the returned values (default int64 if
      jax_enable_x64 is true, otherwise int32).
    key: optional, a PRNG key used as the random key.

  Returns:
    A random array with the specified dtype and with shape given by ``shape`` if
    ``shape is not None, or else by ``lam.shape``."""
  ...

@overload
def rademacher(shape: Shape=(), dtype: DTypeLikeInt=int, key: ArrayLike = None) -> Array: 
  """Sample from a Rademacher distribution.

  The values are distributed according to the probability mass function:

  .. math::
     f(k) = \frac{1}{2}(\delta(k - 1) + \delta(k + 1))

  on the domain :math:`k \in \{-1, 1\}`, where :math:`\delta(x)` is the dirac delta function.

  Args:
    shape: The shape of the returned samples. Default ().
    dtype: The type used for samples.
    key: optional, a PRNG key used as the random key.

  Returns:
    A jnp.array of samples, of shape `shape`. Each element in the output has
    a 50% change of being 1 or -1."""
  ...

@overload
def randint(shape: Shape, minval: IntegerArray, maxval: IntegerArray, dtype: DTypeLikeInt=int, out_sharding: Any=None, key: ArrayLike = None) -> Array: 
  """Sample uniform random values in [minval, maxval) with given shape/dtype.

  Args:
    shape: a tuple of nonnegative integers representing the shape.
    minval: int or array of ints broadcast-compatible with ``shape``, a minimum
      (inclusive) value for the range.
    maxval: int or array of ints broadcast-compatible with ``shape``, a maximum
      (exclusive) value for the range.
    dtype: optional, an int dtype for the returned values (default int64 if
      jax_enable_x64 is true, otherwise int32).
    key: optional, a PRNG key used as the random key.

  Returns:
    A random array with the specified shape and dtype."""
  ...

@overload
def rayleigh(scale: RealArray, shape: Shape | None=None, dtype: DTypeLikeFloat=float, key: ArrayLike = None) -> Array: 
  """Sample Rayleigh random values with given shape and float dtype.

  The values are returned according to the probability density function:

  .. math::
     f(x;\sigma) \propto xe^{-x^2/(2\sigma^2)}

  on the domain :math:`-\infty < x < \infty`, and where :math:`\sigma > 0` is the scale
  parameter of the distribution.

  Args:
    scale: a float or array of floats broadcast-compatible with ``shape``
      representing the parameter of the distribution.
    shape: optional, a tuple of nonnegative integers specifying the result
      shape. Must be broadcast-compatible with ``scale``. The default (None)
      produces a result shape equal to ``scale.shape``.
    dtype: optional, a float dtype for the returned values (default float64 if
      jax_enable_x64 is true, otherwise float32).
    key: optional, a PRNG key used as the random key.

  Returns:
    A random array with the specified dtype and with shape given by ``shape`` if
    ``shape`` is not None, or else by ``scale.shape``."""
  ...

@overload
def split(num: int | tuple[int, ...]=2, key: ArrayLike = None) -> Array: 
  """Splits a PRNG key into `num` new keys by adding a leading axis.

  Args:
    num: optional, a positive integer (or tuple of integers) indicating
      the number (or shape) of keys to produce. Defaults to 2.
    key: optional, a PRNG key (from ``key``, ``split``, ``fold_in``).

  Returns:
    An array-like object of `num` new PRNG keys."""
  ...

@overload
def t(df: RealArray, shape: Shape=(), dtype: DTypeLikeFloat=float, key: ArrayLike = None) -> Array: 
  """Sample Student's t random values with given shape and float dtype.

  The values are distributed according to the probability density function:

  .. math::
     f(t; \nu) \propto \left(1 + \frac{t^2}{\nu}\right)^{-(\nu + 1)/2}

  Where :math:`\nu > 0` is the degrees of freedom, given by the parameter ``df``.

  Args:
    df: a float or array of floats broadcast-compatible with ``shape``
      representing the degrees of freedom parameter of the distribution.
    shape: optional, a tuple of nonnegative integers specifying the result
      shape. Must be broadcast-compatible with ``df``. The default (None)
      produces a result shape equal to ``df.shape``.
    dtype: optional, a float dtype for the returned values (default float64 if
      jax_enable_x64 is true, otherwise float32).
    key: optional, a PRNG key used as the random key.

  Returns:
    A random array with the specified dtype and with shape given by ``shape`` if
    ``shape`` is not None, or else by ``df.shape``."""
  ...

@overload
def triangular(left: RealArray, mode: RealArray, right: RealArray, shape: Shape | None=None, dtype: DTypeLikeFloat=float, key: ArrayLike = None) -> Array: 
  """Sample Triangular random values with given shape and float dtype.

  The values are returned according to the probability density function:

  .. math::
      f(x; a, b, c) = \frac{2}{c-a} \left\{ \begin{array}{ll} \frac{x-a}{b-a} & a \leq x \leq b \\ \frac{c-x}{c-b} & b \leq x \leq c \end{array} \right.

  on the domain :math:`a \leq x \leq c`.

  Args:
    left: a float or array of floats broadcast-compatible with ``shape``
      representing the lower limit parameter of the distribution.
    mode: a float or array of floats broadcast-compatible with ``shape``
      representing the peak value parameter of the distribution, value must
      fulfill the condition ``left <= mode <= right``.
    right: a float or array of floats broadcast-compatible with ``shape``
      representing the upper limit parameter of the distribution, must be
      larger than ``left``.
    shape: optional, a tuple of nonnegative integers specifying the result
      shape. Must be broadcast-compatible with ``left``,``mode`` and ``right``.
      The default (None) produces a result shape equal to ``left.shape``, ``mode.shape``
      and ``right.shape``.
    dtype: optional, a float dtype for the returned values (default float64 if
      jax_enable_x64 is true, otherwise float32).
    key: optional, a PRNG key used as the random key.

  Returns:
    A random array with the specified dtype and with shape given by ``shape`` if
    ``shape`` is not None, or else by ``left.shape``, ``mode.shape`` and ``right.shape``."""
  ...

@overload
def truncated_normal(lower: RealArray, upper: RealArray, shape: Shape | None=None, dtype: DTypeLikeFloat=float, out_sharding: Any=None, key: ArrayLike = None) -> Array: 
  """Sample truncated standard normal random values with given shape and dtype.

  The values are returned according to the probability density function:

  .. math::
     f(x) \propto e^{-x^2/2}

  on the domain :math:`\rm{lower} < x < \rm{upper}`.

  Args:
    lower: a float or array of floats representing the lower bound for
      truncation. Must be broadcast-compatible with ``upper``.
    upper: a float or array of floats representing the  upper bound for
      truncation. Must be broadcast-compatible with ``lower``.
    shape: optional, a tuple of nonnegative integers specifying the result
      shape. Must be broadcast-compatible with ``lower`` and ``upper``. The
      default (None) produces a result shape by broadcasting ``lower`` and
      ``upper``.
    dtype: optional, a float dtype for the returned values (default float64 if
      jax_enable_x64 is true, otherwise float32).
    key: optional, a PRNG key used as the random key.

  Returns:
    A random array with the specified dtype and shape given by ``shape`` if
    ``shape`` is not None, or else by broadcasting ``lower`` and ``upper``.
    Returns values in the open interval ``(lower, upper)``."""
  ...

@overload
def uniform(shape: Shape=(), dtype: DTypeLikeFloat=float, minval: RealArray=0.0, maxval: RealArray=1.0, out_sharding: Any=None, key: ArrayLike = None) -> Array: 
  """Sample uniform random values in [minval, maxval) with given shape/dtype.

  Args:
    shape: optional, a tuple of nonnegative integers representing the result
      shape. Default ().
    dtype: optional, a float dtype for the returned values (default float64 if
      jax_enable_x64 is true, otherwise float32).
    minval: optional, a minimum (inclusive) value broadcast-compatible with shape for the range (default 0).
    maxval: optional, a maximum (exclusive) value broadcast-compatible with shape for the range (default 1).
    key: optional, a PRNG key used as the random key.

  Returns:
    A random array with the specified shape and dtype."""
  ...

@overload
def wald(mean: RealArray, shape: Shape | None=None, dtype: DTypeLikeFloat=float, key: ArrayLike = None) -> Array: 
  """Sample Wald random values with given shape and float dtype.

  The values are returned according to the probability density function:

  .. math::
     f(x;\mu) = \frac{1}{\sqrt{2\pi x^3}} \exp\left(-\frac{(x - \mu)^2}{2\mu^2 x}\right)

  on the domain :math:`-\infty < x < \infty`, and where :math:`\mu > 0` is the location
  parameter of the distribution.


  Args:
    mean: a float or array of floats broadcast-compatible with ``shape``
      representing the mean parameter of the distribution.
    shape: optional, a tuple of nonnegative integers specifying the result
      shape. Must be broadcast-compatible with ``mean``. The default
      (None) produces a result shape equal to ``np.shape(mean)``.
    dtype: optional, a float dtype for the returned values (default float64 if
      jax_enable_x64 is true, otherwise float32).
    key: optional, a PRNG key used as the random key.

  Returns:
    A random array with the specified dtype and with shape given by ``shape`` if
    ``shape`` is not None, or else by ``mean.shape``."""
  ...

@overload
def weibull_min(scale: RealArray, concentration: RealArray, shape: Shape=(), dtype: DTypeLikeFloat=float, key: ArrayLike = None) -> Array: 
  """Sample from a Weibull distribution.

  The values are distributed according to the probability density function:

  .. math::
     f(x;\sigma,c) \propto x^{c - 1} \exp(-(x / \sigma)^c)

  on the domain :math:`0 < x < \infty`, where :math:`c > 0` is the concentration
  parameter, and :math:`\sigma > 0` is the scale parameter.

  Args:
    scale: The scale parameter of the distribution.
    concentration: The concentration parameter of the distribution.
    shape: The shape added to the parameters loc and scale broadcastable shape.
    dtype: The type used for samples.
    key: optional, a PRNG key.

  Returns:
    A jnp.array of samples."""
  ...


__all__ = ['ball', 'bernoulli', 'beta', 'binomial', 'bits', 'categorical', 'cauchy', 'chisquare', 'choice', 'clone', 'dirichlet', 'double_sided_maxwell', 'exponential', 'f', 'fold_in', 'gamma', 'generalized_normal', 'geometric', 'gumbel', 'laplace', 'loggamma', 'logistic', 'lognormal', 'maxwell', 'multinomial', 'multivariate_normal', 'normal', 'orthogonal', 'pareto', 'permutation', 'poisson', 'rademacher', 'randint', 'rayleigh', 'split', 't', 'triangular', 'truncated_normal', 'uniform', 'wald', 'weibull_min']


import jaxpy as jp
import jaxpy.random as random

import jax



def test_simple():
    # Test randn
    # FIXME randn and others in jp
    a=jp.randn((2, 3))
    # Test randint
    a=jp.randint(0, 10, (2, 3))

    a= random.normal((2,3))

    nnx_key = jp.Rngs(0)
    a = random.normal((2,3), key=nnx_key.get_next())

    jax_key = random.key(0)
    a = random.normal((2,3), key=jax_key)

    assert random.normal.__doc__

    # # Test jit with a simple function
    # @jp.jit
    # def simple_function(x):
    #     return x * random.normal((2, 3))

    # print("Testing jit:")
    # print(simple_function(5))

# ## preprocessor tests

# def test_name_error():
#     call()

# def test_name_error2():
#     def test():
#         call()
    
#     test()

# # I consider this success since function is not called
# # should it be????
# def test_name_success():
#     def test():
#         call()

# # the implementation with call would lead to error here...
# # which would be incoherent...
# def test_name_error3():
#     def test():
#         def inner():
#             call()
#         inner()

# # calling on end mainnodeDef would lead to weird successes
# # this would be preprocessed as a success
# def weird_success():
#     def inner1():
#         inner2()
    
#     def innerm():
#         print("Inner m called")

#     inner1()

#     def inner2():
#         print("Inner 2 called")
#         innerm()
    
#     inner1()

# # other problem is redefinitions
# # redefinitions is a hard problem
# def problems_success():
#     def inner1():
#         inner2()
    
#     def innerm():
#         print("Inner m called")
#     def interest_f():
#         print("interest_f called")

#     def inner2():
#         print("Inner 2 called")
#         innerm()

#     inner1()
#     def inner2():
#         print("Inner 2 called")
#         interest_f()

    
#     inner1()


# def test_name_error4():
#     def test():
#         call()
    
#     def inner():
#         test()
    



# test no partial imported, vs user defined partial, key, 
# redefining functions

#FIXME already jitted functions preprocess needs to be done
# FIXME we also need to solve the classes problem
    # init, object construction
    # objects
# (note that jax allows for nnx.Objects classes to be constructed inside jit code)
# annoyances:
# key_arg_name is at good name?

# FIXME even if the user manually sets the keys, we still are modifying
# the function to accept a key argument (as long there is a jaxpy random function),
# and still generating new key values,
# not super eficient, is it worth the change? (benchmarking)

"""
Approach 3.2.1
we are going to use partials for functions of interest,
but also locally reassign and replace all functions of interest names
    with new names

Example
def f():
    a = f2 if something else interest_f
    return a()

# this becomes
def f(key=None):
    new_name = partial(interest_f,key_arg = key)
    a = f2 if something else new_name
    return a()

To allow working with redefinitions of functions and variables, 
we need to realize this only happens on the main funcdef scope.
And that nodeDefs in main function (local variables) can be jit-compiled
on inner nodeDefs, without passing keyword arguments.
This approach should work for all python weird dynamic shenanigans.

3.2.1 change
+ if we modify a function that is in a module, how does it work?
well we compile a new function on that scope. Then we have this new 
function that could be used in other places.
The simplest approach is to create a new variable on the scope
of the upper variable, and use the new name in the modified function

warning nao sei como isto funciona com class methods...
-> penso que existe solucao mas
1. nao deve ser possivel criar classes dentro de jit code
2. we need to preprocess all class args before going to function
(a ideia e' q provavelmente e' registar rng no init, como rng e' global
em principio muda em todo o lado? uma ideia)

question 2:
what if 2 full names represent the same function?
-> the function id will give the same modified function, we will
just also create another variable on the scope (is okay i think)

Basically this means:
1. We only need to modify the main function definition
2. The key argument must be unique from all existent names
in the tree
3. we need to track all names in tree
4. we need detect initial attribute node, assign parent node to it
(simple previous node variable), process full name, set it as full_name on 
node, check if we already processed this node if not process full name (to check if callable and global scope) 
    4.1 if global scope and function of interest, put node on 
local_replace_function set (also if already processed)
5. after we process tree, set functiondef with new key argument,
add assignment nodes that replace function names with partials, 
go to all nodes in local_replace_function set, go to parent,
and replace name with replaced name

Implementation details:
+ To get all names in tree, we need to traverse the entire graph 
first. On the second traversal we only need a few nodes, so we 
can actually just same the important nodes to edit later
+ So in first traversal, what variables should we track?
    + first functiondef node, this will be the main function
    + all used names (not attribute, just names) in the tree

+ create a visit method that before calling generic_visit,
sets self.parent = node before visiting children
    
+ if name node, always add name to the set of names 
    (this encompasses) all names including calls, etc.
    + also do other part of attr node logic

+ if attr node, create mutex like thing, that 
sets true if inside_attr_chain. If name node just run follow thing
    + if inside_attr_chain is True, then do nothing/add name
    + if inside_attr_chain is False, then
        process full name, then check if in full_name_to_node_dict.keys():
        if not in full_name_to_node_dict.keys(),
            then check full name scope,
            if not in local names (current)
            and is global and callable, 
            (check first if already processed in global_ids_processed)
            if not process callable, do everything, including setting modified
                function in the scope

        if it is function of interest or in full_name_to_node_dict.keys():
            set full_name on node,
            set node.parent = self.parent
            add node to replace_function_node list
            add full_name: node to full_name_to_node_dict (dont re add if already in keys)
            set self.modified = True

+ if functiondef node, not main, add name to the set of names, 
add arguments and keywords also to set of names

Second traversal/ Final Processing:
in the main FunctionDef node, after visiting the entire tree
+ if class.modified is True, then we need to modify the function
+ find a unique key argument name, that is not in the set of names
+ add the key argument to the function definition

+ for each full_name in replace_function_dict:
    + create a new_unique name
    + append to functionDef assignment node with name new_unique_name
    and value replace_function_dict[full_name] (the name/attr node)
    + add full_name: new_unique_name to prev_name_to_new_name_dict
+ for each node in replace_function_node,
    + go to node.parent (fuck where to look???, 
    we need to find object with same id as node, search all attributes?
    seems a good approach with ast.iter_fields(node), gives gen dict attr:value
    do for attr, value, with raise if did not found node)
    + replace object with ast.Name(id=prev_name_to_new_name_dict[node.full_name], ctx=ast.Load())
+ end TransformAstTransformer class
+ finish the process function with the new compiled function

"""

from functools import partial
import uuid
import ast
import inspect
import textwrap

def resolve_attr_chain(node, scopes):
    """
    node: ast.Node
    scopes: (global_scope, local_scope)
    returns: (variable, full_name, if is in local scope)
    """
    if isinstance(node, ast.Name):
        global_scope, local_scope = scopes

        if local_scope and node.id in local_scope:
            return local_scope[node.id], node.id, True
        
        return global_scope.get(node.id), node.id, False
    elif isinstance(node, ast.Attribute):
        base, full_name, is_local = resolve_attr_chain(node.value, scopes)
        if base:
            return getattr(base, node.attr, None), f"{full_name}.{node.attr}", is_local
        
    elif isinstance(node, ast.Call):
        # then we are calling within an attribute
        # like a().b()
        # we will not resolve this #FIXME
        # FIXME we should actually still process the call.....
        """
        Example here there are 2 calls
....
                                value=Call(
                                    func=Attribute(
                                        value=Call(
                                            func=Attribute(
                                                value=Name(id='jp', ctx=Load()),
                                                attr='Rngs',
                                                ctx=Load()),
                                            args=[
                                                Constant(value=20)],
                                            keywords=[]),
                                        attr='get_next',
                                        ctx=Load()),
....
If not processing the attribute (as it is not defined in global scope),
we should process at least the inner for Rngs
        
        """
        return None, None, None
    else:
        raise ValueError(f"Node {node} is not expected, (Name, Attribute or Call expected).")
    return None, full_name, None

def get_all_args(a: ast.arguments):
    args = []
    args += a.args if isinstance(a.args, list) else [a.args]
    args += a.vararg if isinstance(a.vararg, list) else [a.vararg]
    args += a.kwonlyargs if isinstance(a.kwonlyargs, list) else [a.kwonlyargs]
    args += a.kwarg if isinstance(a.kwarg, list) else [a.kwarg]
    return [arg.arg for arg in args if isinstance(arg, ast.arg)]


# preprocess cache
global_ids_processed = {} # stores the info of processed functions, important to retrieve key name
global_id_to_function = {} 




class SomethingTransformer(ast.NodeTransformer):
    """

    """



    def __init__(self, global_scope, local_scope):
        super().__init__()
        self.global_scope = global_scope
        self.local_scope = local_scope
        # FIXME precisamos de arg para saber se tree foi modificada ou nao, mas ja temos o info???
        self.modified = False #?
        # FIXME e saber qual a key name....
        self.info = {} #?

        self._did_we_process_main_funcDef = False
        self._names = set() # set of all names in the tree
        self._replace_function_node_list = []
        self._full_name_to_node_dict = {} # fullname: attr/name node
        self._prev_name_to_new_name_dict = {}
        self._curr_inside_attr_chain = False # mutex to check if we are inside an attr chain
        self.parent_stack = [] # keep track of parent nodes
        self._modified_scope_names = {} # for new functions that require patching, we will store the new names here


    
    def generic_visit(self, node):
        self.parent_stack.append(node)
        new_node = super().generic_visit(node)
        self.parent_stack.pop()
        return new_node

    # def get_unique_key_arg_name(self, func_def_node: ast.FunctionDef):
    #     """
    #     Generate a unique key argument name for the function definition node.
    #     This is used to ensure that the key argument does not conflict with existing arguments.
    #     """
    #     base_name = "key"
    #     existing_names = {arg.arg for arg in func_def_node.args.args}
    #     unique_name = base_name
    #     counter = 1
    #     while unique_name in existing_names:
    #         unique_name = f"{base_name}_{counter}"
    #         counter += 1
    #     return unique_name
    
    def get_unique_name(self, base_name=""):
        unique_name = f"{base_name}_{uuid.uuid4().hex}"
        while unique_name in self._names:
            unique_name = f"{base_name}_{uuid.uuid4().hex}"

        self._names.add(unique_name)
        return unique_name

    def _import_partial_to_scope(self):
        """
        Ensure that the 'partial' function from functools is imported in the scope.
        This is necessary for the compilation stage to work correctly.

        To avoid user defining a variable called 'partial' in the scope,
        we will use a unique name for the partial function.
        """
        unique_partial_name = self.get_unique_name(base_name="partial")
        self.global_scope[unique_partial_name] = partial

        self._partial_name = unique_partial_name



    def visit_FunctionDef(self, node):
        if self._did_we_process_main_funcDef:
            self._names.add(node.name)
            return self.generic_visit(node) # just process the node normally

        # this section will only run on the main function definition
        self._did_we_process_main_funcDef = True 

        # add all scope names to the set of names
        self._names.add(node.name)
        self._names.update(get_all_args(node.args))
        self._names.update(self.global_scope.keys())
        self._names.update(self.local_scope.keys())

        node = self.generic_visit(node)

        if self.modified:
            # then this function definition as functions that require a key argument
            # we start by finding a unique key argument, not used anywhere in the scope/tree
            key_arg_name = self.get_unique_name(base_name="key")

            self.info['key_arg_name'] = key_arg_name # store the key name in the info dict

            # add the key argument to the function definition
            node.args.args.append(ast.arg(arg=key_arg_name, annotation=None))
            node.args.defaults.append(ast.Constant(value=None))  # requires Python 3.8+

            # Ensure 'partial' is available in the scope
            self._import_partial_to_scope()  

            # go through all full names
            for full_name, fname_node in self._full_name_to_node_dict.items():
                # we will create a new local variable in the function
                # that does partial(func, key=key_arg_name.get_next())
                new_unique_name = self.get_unique_name(base_name=full_name.split('.')[-1])


                # should use the new unique name in scope, or previous name?
                if full_name in self._modified_scope_names.keys():
                    func_node = ast.Name(id=self._modified_scope_names[full_name], ctx=ast.Load())
                else:
                    func_node = fname_node # we use the previous attr/name node
    
                key_value_node = ast.Name(id=key_arg_name, ctx=ast.Load())

                if fname_node.jaxpy_random_function:
                    # if it is a jaxpy random function, we should get next key
                    # key() # is the same as key.get_next() 
                    key_value_node = ast.Call(
                        func=key_value_node,
                        args=[],
                        keywords=[]
                    )

                # new_name = partial(func, key=key_arg_name.get_next())
                assignment_node = ast.Assign(
                    targets=[ast.Name(id=new_unique_name, ctx=ast.Store())],
                    value=ast.Call(
                        func=ast.Name(id=self._partial_name, ctx=ast.Load()),
                        args=[func_node],
                        keywords=[ast.keyword(arg=fname_node.key_arg_name, value=key_value_node)]
                    )
                )

                # add the assignment node to the function definition
                node.body.insert(0, assignment_node)

                # keep track of the previous name to new name mapping
                self._prev_name_to_new_name_dict[full_name] = new_unique_name
            
            # now we can replace all function references in the tree
            for fname_node in self._replace_function_node_list: # get all name/attr nodes that need to be replaced
                parent = fname_node.parent

                # find the attribute or name node in the parent
                for attr, value in ast.iter_fields(parent):
                    if value is fname_node: 
                        # replace the value with a new Name node with the new unique name
                        setattr(parent, attr, ast.Name(id=self._prev_name_to_new_name_dict[fname_node.full_name], ctx=ast.Load()))
                        break

        return node

    def visit_Name(self, node):
        """
        Visit a Name node and add its id to the set of names.
        If it is an attribute, we will process it later.
        """
        self._names.add(node.id)

        # if we are inside an attr chain, do not process this name
        if self._curr_inside_attr_chain:
            return self.generic_visit(node)

        self._process_node_name(node)

        return self.generic_visit(node)

    def visit_Attribute(self, node):
        if self._curr_inside_attr_chain:
            return self.generic_visit(node)

        # Set the mutex to True, we are now inside an attr chain
        self._curr_inside_attr_chain = True

        self._process_node_name(node)

        node = self.generic_visit(node)

        # After visiting the node, set the mutex back to False
        self._curr_inside_attr_chain = False

        return node



    def _process_node_name(self,node):
        """
        Processes the Attribute or Name node to resolve its full name
        and check if it is a function of interest.
        """
        # Process the full name
        func, full_name, is_local = resolve_attr_chain(node, (self.global_scope, self.local_scope))

        if func is None or not inspect.isfunction(func):
            # FIXME we are not processing classes created inside the function
            # then is a local or not a callable
            return

        # if func in global_id_to_function FIXME
        # Process the called function
        key_arg_name = None
        if hasattr(func, '__jaxpy_random_function__'):
            print(f"Function {func.__name__} is marked as a JAX random function.")
            key_arg_name = 'key' # By default, __jaxpy_random_function__ use key argument

            # This could actually be also a __jaxpy_key_arg_name__ variable if interesting to have
            # but will always need to be a keyword, so that positional argument does set the key
            # and such we can search here directly for keyword arguments

        else:
            # check if we need to process the function
            if id(func) not in global_ids_processed.keys():
                print(f"Processing function {func.__name__}")
                # Processes new tree
                # if function is defined in previous function local scope
                # this means it at same level as the current tree,
                # thus we also need to pass the local scope to process this function
                local_scope = self.local_scope if is_local else {} 
                modified_func, info = preprocess_function(func, local_scope=local_scope)

                # FIXME we should check if is truly modified

                # this new function is not defined in this scope
                # so we need to create a new name not in the scope to use
                # this global value should also not be a functiondef args/keywarg
                # (since we are going to assign to a new variable in the beggining of functiondef, 
                # does not need to be different than follow up variables)

                if id(func) != id(modified_func):
                    # then we have a new function that is not in the scope of this function
                    new_unique_name = self.get_unique_name(base_name=func.__name__)
                    self.global_scope[new_unique_name] = modified_func
                    self._modified_scope_names[full_name] = new_unique_name

            else:
                # already processed just get info
                info = global_ids_processed[id(func)]

                if id(func) in global_id_to_function.keys(): # id(func) != id(modified_func)
                    # then we have a modified function that might not be in the scope
                    if full_name not in self._modified_scope_names.keys():
                        # then the modification is not yet in the scope
                        new_unique_name = self.get_unique_name(base_name=func.__name__)
                        self.global_scope[new_unique_name] = global_id_to_function[id(func)]
                        self._modified_scope_names[full_name] = new_unique_name
                    
                    # otherwise, it is already processed and already in the scope

            if 'key_arg_name' in info:
                # then function was modified to accept a key
                key_arg_name = info['key_arg_name']

        if key_arg_name is not None:
            # add details for future processing
            self._replace_function_node_list.append(node)
            node.parent = self.parent_stack[-1]
            node.full_name = full_name
            node.key_arg_name = key_arg_name 
            node.jaxpy_random_function = hasattr(func, '__jaxpy_random_function__') # either to call .get_next() or not
            self._full_name_to_node_dict[full_name] = node # needed for going through each full_name later
            self.modified = True


# FIXME local_scope, where is going to be used? 1st call should be None?
# this is not a public interface
def preprocess_function(func, local_scope=None):
    """
    FIXME it needs to also change scope with function calls that were modified

    This will also be called everytime we find a function call, while processing the AST.

    returns: 
    + compiled function
    + info dictionary, currently will have key_arg_name if it was added
    """
    # FIXME we probably need more info like key argument name.
    source = inspect.getsource(func)
    dedented_source = textwrap.dedent(source) # remove leading whitespace, specially inside functionDefs
    tree = ast.parse(dedented_source)

    print("Textwrap")
    print(source)
    print(dedented_source)
    # Create a transformer instance
    global_scope = func.__globals__.copy()

    if local_scope is None:
        # this is necessary to obtain local functiondefs from where preprocessed is called
        # WARNING: if used in jit code, this should be done in jit function... not here
        local_scope = inspect.currentframe().f_back.f_locals


    transformer = SomethingTransformer(global_scope, local_scope)

    # Transform the AST
    transformed_tree = transformer.visit(tree)
    ast.fix_missing_locations(transformed_tree)

    

    if transformer.modified:
        # DEBUG
        source_code = ast.unparse(transformed_tree)
        print(f"Modified source code for {func.__name__}:\n")
        print(source_code)
        print("\n")

        # merge scopes
        global_scope.update(local_scope)

        # Compile the modified AST back to code
        compiled = compile(transformed_tree, filename="<ast>", mode="exec")
        exec(compiled, global_scope)

        modified_function = global_scope[func.__name__] # in the function scope will always just be a name
        global_id_to_function[id(func)] = modified_function
        global_ids_processed[id(func)]=transformer.info

        return modified_function, transformer.info
    else:
        # If no modifications were made, return the original function and an empty info dict
        global_ids_processed[id(func)] = {}
        return func, {}
